
/* Interchange File Format */

#include "IFF.h"

#include <stdio.h>
#include <stdlib.h>

#include <dos/dos.h>
#include <libraries/iffparse.h>
#include <datatypes/pictureclass.h>
#include <exec/memory.h>
#include <clib/exec_protos.h>
#include <clib/dos_protos.h>
#include <clib/iffparse_protos.h>
#include <clib/graphics_protos.h>

typedef struct StoredProperty StoredProperty;
typedef struct ContextNode ContextNode;

static void closeStream(ULONG s, BOOL clip)
{
    clip ? CloseClipboard((struct ClipboardHandle *)s) : Close((BPTR)s);
}

IFFHandle *openIFF(STRPTR name, BOOL write, BOOL *clipOut, LONG *errOut)
{
    IFFHandle *iff;
    BOOL clip;
    UBYTE clipUnit;
    LONG err;

    if ((iff = AllocIFF()) == NULL)
    {
        err = NOMEM;
        puts("Out of free memory!");
    }
    else
    {
        if (clip = (name[0] == '-' && name[1] == 'c'))
        {
            clipUnit = name[2] ? atoi(name + 2) : PRIMARY_CLIP;
            if (!(iff->iff_Stream = (ULONG)OpenClipboard(clipUnit)))
                printf("Couldn't open clipboard unit %d!\n", clipUnit);
            else
            {
                InitIFFasClip(iff);
            }
        }
        else
        {
            if (!(iff->iff_Stream = (ULONG)Open(name, write ? MODE_NEWFILE : MODE_OLDFILE)))
                printf("Couldn't open file %s for %s!\n", name, write ? "writing" : "reading");
            else
            {
                InitIFFasDOS(iff);
            }		
        }
        if (iff->iff_Stream == 0)
            err = NOFILE;
        else
        {
            if ((err = OpenIFF(iff, write ? IFFF_WRITE : IFFF_READ)) != 0)
                printf("OpenIFF failed returncode %d!\n", err);
            else
            {
                *clipOut = clip;
                return(iff);
            }			
            closeStream(iff->iff_Stream, clip);
        }
        FreeIFF(iff);
    }
    *errOut = err;
    return(NULL);
}

void closeIFF(IFFHandle *iff, BOOL clip)
{
    CloseIFF(iff);
    closeStream(iff->iff_Stream, clip);
    FreeIFF(iff);
}

LONG parseILBM(IFFHandle *iff, BitMapHeader *bmhdOut)
{
    LONG err;
    LONG props[] =
    {
        ID_ILBM, ID_BMHD,
        ID_ILBM, ID_CAMG,
        ID_ILBM, ID_CMAP
    };
    StoredProperty *sp;

    if ((err = PropChunks(iff, props, 3)) != 0)
        printf("PropChunks failed returncode %d!\n", err);
    else
    {
        if ((err = StopChunk(iff, ID_ILBM, ID_BODY)) != 0)
            printf("StopChunk failed returncode %d!\n", err);
        else
        {
            if ((err = ParseIFF(iff, IFFPARSE_SCAN)) != 0 && err != IFFERR_EOC && err != IFFERR_EOF)
                printf("ParseIFF failed returncode %d!\n", err);
            else
            {
                if ((sp = FindProp(iff, ID_ILBM, ID_BMHD)) == NULL)
                    printf("Couldn't find BMHD chunk!\n");
                else
                {
                    *bmhdOut = *(BitMapHeader *)sp->sp_Data;
                    return(err);
                }
            }
        }
    }
    return(err);
}

BOOL findCAMG(IFFHandle *iff, ULONG *camgOut)
{
    StoredProperty *sp;

    if ((sp = FindProp(iff, ID_ILBM, ID_CAMG)) == NULL)
        printf("Warning: Couldn't find CAMG chunk.\n");
    else
    {
        *camgOut = *(ULONG *)sp->sp_Data;
        return(TRUE);
    }
    return(FALSE);
}

BOOL findCMAP(IFFHandle *iff, ULONG **rgb32Out)
{
    StoredProperty *sp;
    UBYTE *data;
    LONG size;
    ULONG *rgb32;
    WORD i;

    if ((sp = FindProp(iff, ID_ILBM, ID_CMAP)) == NULL)
        printf("Couldn't find CMAP chunk!\n");
    else
    {
        data = (UBYTE *)sp->sp_Data;
        size = sp->sp_Size;

        if ((rgb32 = (ULONG *)AllocVec((size + 2) * sizeof(ULONG), MEMF_PUBLIC)) == NULL)
            printf("Out of free memory!\n");
        else
        {
            rgb32[0] = (size / 3) << 16; /* Color count */
            for (i = 1; i <= size; i++)
            {
                UBYTE color = *data++;
                rgb32[i] = RGB(color);
            }
            rgb32[size + 1] = 0L;
            *rgb32Out = rgb32;
            return(TRUE);
        }
    }
    return(FALSE);
}

static BOOL unpackRow(BYTE **bufPtr, LONG *sizePtr, BYTE *plane, WORD bpr, UBYTE cmp)
{
    BYTE *buf = *bufPtr;
    LONG size = *sizePtr;
    BYTE con;
    WORD count;
    BYTE data;

    if (cmp == cmpNone)
    {
        if (size < bpr)
        {
            printf("Buffer underflow!\n");
            return(FALSE);
        }
        CopyMem(buf, plane, bpr);
        buf += bpr;
        size -= bpr;
    }
    else if (cmp == cmpByteRun1)
    {
        while (bpr > 0)
        {
            if (size < 1)
            {
                printf("Buffer underflow!\n");
                return(FALSE);
            }
            size--;
            if ((con = *buf++) >= 0)
            {
                count = con + 1;
                if (size < count || bpr < count)
                {
                    printf("Buffer under/overflow!\n");
                    return(FALSE);
                }
                size -= count;
                bpr -= count;
                while (count-- > 0)
                {
                    *plane++ = *buf++;
                }
            }
            else if (con != -128)
            {
                count = (-con) + 1;
                if (size < 1 || bpr < count)
                {
                    printf("Buffer under/overflow!\n");
                    return(FALSE);
                }
                size--;
                bpr -= count;
                data = *buf++;
                while (count-- > 0)
                {
                    *plane++ = data;
                }
            }
        }
    }
    else
    {
        printf("Unknown compression!\n");
        return(FALSE);
    }
    *bufPtr = buf;
    *sizePtr = size;
    return(TRUE);
}

BOOL readILBM(IFFHandle *iff, BitMapHeader *bmhd, BitMap **bmOut)
{
    ContextNode *cn;
    BYTE *buffer, *curbuf;
    LONG size, cursize;
    BitMap *bm;
    WORD width = bmhd->bmh_Width;
    WORD height = bmhd->bmh_Height;
    UBYTE depth = bmhd->bmh_Depth;
    UBYTE msk = bmhd->bmh_Masking;
    UBYTE cmp = bmhd->bmh_Compression;
    WORD row;
    UBYTE plane;
    PLANEPTR planes[8 + 1];
    WORD bpr, mod;
    BOOL success = FALSE;

    if (msk == mskHasMask)
    {
        depth++;
    }
    
    if ((cn = CurrentChunk(iff)) == NULL)
        printf("Couldn't get current chunk!\n");
    else
    {
        size = cn->cn_Size;
        if ((buffer = (BYTE *)AllocMem(size, MEMF_PUBLIC)) == NULL)
            printf("Out of free memory!\n");
        else
        {
            if (ReadChunkBytes(iff, buffer, size) != size)
                printf("Couldn't read chunk bytes!\n");
            else
            {
                if ((bm = AllocBitMap(width, height, depth, 0, NULL)) == NULL)
                    printf("Out of free graphics memory!\n");
                else
                {
                    mod = bm->BytesPerRow;
                    bpr = RowBytes(width);
                    curbuf = buffer;
                    cursize = size;

                    for (plane = 0; plane < depth; plane++)
                    {
                        planes[plane] = bm->Planes[plane];
                    }

                    for (row = 0; row < height; row++)
                    {
                        for (plane = 0; plane < depth; plane++)
                        {
                            if (!(success = unpackRow(&curbuf, &cursize, planes[plane], bpr, cmp)))
                                break;
                            planes[plane] += mod;
                        }
                        if (!success)
                            break;
                    }
                    if (success)
                    {
                        FreeMem(buffer, size);
                        *bmOut = bm;
                        return(TRUE);
                    }
                    FreeBitMap(bm);
                }
            }
            FreeMem(buffer, size);
        }
    }
    return(FALSE);
}
