
#include <intuition/screens.h>
#include <graphics/gfxmacros.h>
#include <exec/memory.h>
#include <exec/interrupts.h>
#include <hardware/intbits.h>

#include <clib/exec_protos.h>
#include <clib/graphics_protos.h>
#include <clib/intuition_protos.h>

#include "Screen.h"

#ifndef AMIGA
#define __far
#define __interrupt
#define __asm
#define __a1
#endif

__far extern struct Custom custom;

/* Obtain best screen mode for given resolution, colors and public screen monitor */

ULONG bestModeID(WORD width, WORD height, UBYTE depth, STRPTR pubName)
{
	struct Screen *pubs;

	if (pubs = LockPubScreen(pubName))
	{
		struct ViewPort *vp = &pubs->ViewPort;
		ULONG pubMode;

		UnlockPubScreen(pubName, pubs);

		if ((pubMode = GetVPModeID(vp)) != INVALID_ID)
		{
			ULONG pubMon = pubMode & MONITOR_ID_MASK;
			ULONG modeID;

			if ((modeID = BestModeID(
				BIDTAG_NominalWidth, width,
				BIDTAG_NominalHeight, height,
				BIDTAG_Depth, depth,
				BIDTAG_MonitorID, pubMon,
				TAG_DONE)) != INVALID_ID)
			{
				return(modeID);
			}
		}
	}
	return(INVALID_ID);
}

/* Obtain display clip for given screen mode */

BOOL getDClip(ULONG modeID, struct Rectangle *saveDClip)
{
	DisplayInfoHandle dih;

	if (dih = FindDisplayInfo(modeID))
	{
		struct DimensionInfo dims;

		if (GetDisplayInfoData(dih, (UBYTE *)&dims, sizeof(dims), DTAG_DIMS, modeID) > 0)
		{
			*saveDClip = dims.Nominal;
			return(TRUE);
		}
	}
	return(FALSE);
}

/* Allocate screen bitmap for given display clip and colors */

/* Fill this bitmap before displaying */

struct BitMap *allocBitMap(struct Rectangle *clip, UBYTE depth)
{
	struct BitMap *bm;

	if (bm = AllocBitMap(clip->MaxX - clip->MinX + 1, clip->MaxY - clip->MinY + 1, depth, BMF_DISPLAYABLE | BMF_INTERLEAVED, NULL))
	{
		return(bm);
	}
	return(NULL);
}

/* Open exclusive screen using modeID, display clip and bitmap and optional tags */

struct Screen *openScreen(ULONG modeID, struct Rectangle *clip, struct BitMap *bm, ULONG tag, ...)
{
	struct Screen *s;

	if (s = OpenScreenTags(NULL,
		SA_DisplayID, modeID,
		SA_DClip, clip,
		SA_BitMap, bm,
		SA_Exclusive, TRUE,
		TAG_MORE, &tag))
	{
		return(s);
	}
	return(NULL);
}

/* Copper interrupt server invoked at display line 0. 
 * Signal our task when our ViewPort is visible */

__asm __interrupt LONG myCopServer(register __a1 struct copData *cd)
{
	if (!(cd->vport->Modes & VP_HIDE))
	{
		Signal(cd->sigTask, 1L << cd->sigBit);
	}
	return(0);
}

/* Attach user graphics coprocessor list to the screen to receive its interrupt signal
 * when display beam is at line 0 (and optionally install other commands like rainbow etc.) */

struct Interrupt *attachUserCopList(struct Screen *s, UWORD commands, void (*addCommands)(struct Screen *s, struct UCopList *ucl), BYTE pri)
{
	struct Interrupt *is;

	if (is = AllocMem(sizeof(*is), MEMF_PUBLIC | MEMF_CLEAR))
	{
		struct copData *cd;

		if (cd = AllocMem(sizeof(*cd), MEMF_PUBLIC | MEMF_CLEAR))
		{
			struct UCopList *ucl;

			is->is_Code = (void(*)())myCopServer;
			is->is_Data = (APTR)cd;
			is->is_Node.ln_Pri = pri;
			is->is_Node.ln_Name = "GameX";

			if ((cd->sigBit = AllocSignal(-1)) != -1)
			{
				cd->sigTask = FindTask(NULL);
				cd->vport = &s->ViewPort;

				if (ucl = AllocMem(sizeof(*ucl), MEMF_PUBLIC | MEMF_CLEAR))
				{
					CINIT(ucl, commands);
					CWAIT(ucl, 0, 0);
					CMOVE(ucl, custom.intreq, INTF_SETCLR | INTF_COPER);
					if (addCommands)
					{
						(*addCommands)(s, ucl);
					}
					CEND(ucl);

					Forbid();
					s->ViewPort.UCopIns = ucl;
					Permit();

					RethinkDisplay();

					AddIntServer(INTB_COPER, is);
					return(is);
				}
				FreeSignal(cd->sigBit);
			}
			FreeMem(cd, sizeof(*cd));
		}
		FreeMem(is, sizeof(*is));
	}
	return(NULL);
}

/* Free copper data */

void detachUserCopList(struct Interrupt *is)
{
	struct copData *cd = (struct copData *)is->is_Data;

	RemIntServer(INTB_COPER, is);
	FreeSignal(cd->sigBit);
	FreeMem(cd, sizeof(*cd));
	FreeMem(is, sizeof(*is));
}

/* Allocate screen buffers and safe message port */

BOOL allocScreenBuffers(struct Screen *s, struct BitMap *bm[2], struct dbufData *dd)
{
	if (dd->safePort = CreateMsgPort())
	{
		if (dd->buffers[0] = AllocScreenBuffer(s, bm[0], 0))
		{
			if (dd->buffers[1] = AllocScreenBuffer(s, bm[1], 0))
			{
				dd->safeToDraw = TRUE;
				dd->toggleFrame = 1;
				dd->held = FALSE;

				dd->buffers[0]->sb_DBufInfo->dbi_SafeMessage.mn_ReplyPort = dd->safePort;
				dd->buffers[1]->sb_DBufInfo->dbi_SafeMessage.mn_ReplyPort = dd->safePort;

				return(TRUE);
			}
			FreeScreenBuffer(s, dd->buffers[0]);
		}
		DeleteMsgPort(dd->safePort);
	}
	return(FALSE);
}

/* Free screen buffers */

void freeScreenBuffers(struct Screen *s, struct dbufData *dd)
{
	if (!dd->safeToDraw)
	{
		while (!GetMsg(dd->safePort))
		{
			Wait(1L << dd->safePort->mp_SigBit);
		}
	}
	FreeScreenBuffer(s, dd->buffers[1]);
	FreeScreenBuffer(s, dd->buffers[0]);
	DeleteMsgPort(dd->safePort);
}

/* Make sure we can draw into other buffer */

void safeToDraw(struct dbufData *dd, struct RastPort *rp)
{
	if (!dd->safeToDraw)
	{
		while (!GetMsg(dd->safePort))
		{
			Wait(1L << dd->safePort->mp_SigBit);
		}
		dd->safeToDraw = TRUE;
		rp->BitMap = dd->buffers[dd->toggleFrame];
	}
}

/* Change buffers if ready (call upon signal from copper interrupt) */

BOOL changeScreenBuffer(struct Screen *s, struct dbufData *dd)
{
	if (dd->safeToDraw)
	{
		dd->held = !ChangeScreenBuffer(s, dd->buffers[dd->toggleFrame]);

		if (!dd->held)
		{
			dd->toggleFrame ^= 1;
			dd->safeToDraw = FALSE;
			return(TRUE);
		}
	}
	return(FALSE);
}
